import numpy as np
from _tsadeval.metrics import Binary_anomalies, pointwise_to_full_series, segmentwise_to_full_series, DelayThresholdedPointAdjust
def get_tp_tn_fp_fn_point_wise(y_true: np.array,y_pred: np.array):
    TP,TN,FP,FN=0,0,0,0
    for true,pred in zip(y_true,y_pred):
        if true==pred:
            if true==1:
                TP+=1
            else:
                TN+=1
        else:
            if true==1:
                FN+=1
            else:
                FP+=1
    return TP,TN,FP,FN


def get_events(y_true,anomaly=True):
    events = []
    start_idx = None
    v = 0
    if anomaly:
        v = 1
    else:
        v = 0

    for i, val in enumerate(y_true):
        if val == v:  # Si encontramos el inicio de un evento
            if start_idx is None:
                start_idx = i  # Establecemos el inicio del evento
        elif start_idx is not None:  # Si encontramos el final de un evento
            events.append((start_idx, i - 1))  # Agregamos el evento a la lista de eventos
            start_idx = None  # Restablecemos el inicio del evento

    if start_idx is not None:  # Si al final de la secuencia aún estamos dentro de un evento
        events.append((start_idx, len(y_true) - 1))  # Agregamos el evento final a la lista de eventos


    return events

def calculate_intersection(event1, event2):
    start_intersection = max(event1[0], event2[0])
    end_intersection = min(event1[1], event2[1])

    # If there is an intersection, return the range of the intersection, otherwise return None
    if start_intersection <= end_intersection:
        return [start_intersection, end_intersection]
    else:
        return None

def get_tp_tn_fp_fn_point_adjusted(y_true: np.array,y_pred: np.array):
    TP, TN, FP, FN = get_tp_tn_fp_fn_point_wise(y_true, y_pred)
    TP=0
    FN=0
    y_true_events = get_events(y_true,anomaly=True)
    y_pred_events = get_events(y_pred,anomaly=True)

    i_true = 0
    i_pred = 0
    while i_true<len(y_true_events):
        detected = False
        while i_pred<len(y_pred_events) and y_true_events[i_true][1]>y_pred_events[i_pred][0]:
            if calculate_intersection(y_true_events[i_true],y_pred_events[i_pred]) is not None:
                TP+= y_true_events[i_true][1]-y_true_events[i_true][0]+1
                detected=True
                break
            elif y_true_events[i_true][0]>y_pred_events[i_pred][1]:
                i_pred+=1
        
        if not detected:
            FN+= y_true_events[i_true][1]-y_true_events[i_true][0]+1
        i_true+=1
    
    return TP, TN, FP, FN

def get_tp_tn_fp_fn_delay_th_point_adjusted(y_true: np.array,y_pred: np.array,k: int):
    TP, TN, FP, FN = get_tp_tn_fp_fn_point_wise(y_true, y_pred)
    TP=0
    FN=0
    y_true_events = get_events(y_true,anomaly=True)
    y_pred_events = get_events(y_pred,anomaly=True)

    i_true = 0
    i_pred = 0
    while i_true<len(y_true_events):
        detected = False
        while i_pred<len(y_pred_events) and y_true_events[i_true][1]>y_pred_events[i_pred][0]:
            intersec = calculate_intersection(y_true_events[i_true],y_pred_events[i_pred])
            if intersec is not None and intersec[0]-y_true_events[i_true][0]<k:
                TP+= y_true_events[i_true][1]-y_true_events[i_true][0]+1
                detected=True
                break
            else:
                i_pred+=1
        
        if not detected:
            FN+= y_true_events[i_true][1]-y_true_events[i_true][0]+1
        i_true+=1
    
    return TP, TN, FP, FN

def get_tp_tn_fp_fn_point_adjusted_at_k(y_true: np.array,y_pred: np.array, k: float):
    TP, TN, FP, FN = get_tp_tn_fp_fn_point_wise(y_true, y_pred)
    TP=0
    FN=0
    y_true_events = get_events(y_true,anomaly=True)
    y_pred_events = get_events(y_pred,anomaly=True)

    i_true = 0
    i_pred = 0
    while i_true<len(y_true_events):
        detected = False
        while i_pred<len(y_pred_events) and y_true_events[i_true][1]>y_pred_events[i_pred][0]:
            intersec = calculate_intersection(y_true_events[i_true],y_pred_events[i_pred])
            if intersec is not None:
                event_size = y_true_events[i_true][1]-y_true_events[i_true][0]+1
                intersec_size = intersec[1]-intersec[0]+1
            if intersec is not None and intersec_size/event_size>=k:
                
                TP+= y_true_events[i_true][1]-y_true_events[i_true][0]+1
                detected=True
                break
            else:
                i_pred+=1
        
        if not detected:
            FN+= y_true_events[i_true][1]-y_true_events[i_true][0]+1
        i_true+=1
    
    return TP, TN, FP, FN


def get_tp_tn_fp_fn_latency_sparsity_aw(y_true: np.array, y_pred: np.array, ni: int):
    batched_shape = (int(np.ceil(y_pred.shape[0] / ni)), 1)
    label_batch = np.zeros(batched_shape)
    pred_batch = np.zeros(batched_shape)
    actual = np.copy(y_true)
    predict = np.copy(y_pred)
    detect_state = False  # triggered when a True anomaly is detected by model
    anomaly_batch_count = 0
    i, i_ni = 0, 0
    step = ni

    while i < len(y_true) and step > 1:
        j = min(i + step, len(y_true))  # end of ni (batch) starting at i

        # Adjust step size if needed
        if step > 2 and actual[i:j].sum() > 1:
            if np.diff(np.where(actual[i:j])).max() > 1:  # if it finds an interruption in the true label continuity
                step = min(int((j - i) / 2), 2)  # reduce step size
                label_batch = np.append(label_batch, [[0]], axis=0)
                pred_batch = np.append(pred_batch, [[0]], axis=0)  # increase size
                j = i + step
            else:
                step = ni
        else:
            step = ni

        # Start rolling window scoring
        if actual[i:j].max():  # If label = T
            if not actual[i]:  # if first value is normal
                detect_state = False
            s = actual[i:j].argmax()  # this is the index of the first occurrence
            if detect_state:  # if anomaly was previously detected by model
                anomaly_batch_count += 1
                pred_batch[i_ni], label_batch[i_ni], predict[i + s:j] = 1, 1, 1
            elif predict[i:j].max():  # if alert was detected with T
                detect_state = True  # turn on detection state
                anomaly_batch_count += 1
                pred_batch[i_ni], label_batch[i_ni], predict[i + s:j] = 1, 1, 1
            else:
                detect_state = False
                label_batch[i_ni] = 1
        else:
            detect_state = False
            if predict[i:j].max():  # if False positive
                pred_batch[i_ni] = 1
        i += step
        i_ni += 1

    if ni == 1:
        return get_tp_tn_fp_fn_point_wise(actual, predict)

    return get_tp_tn_fp_fn_point_wise(label_batch.flatten().astype(int), pred_batch.flatten().astype(int))

def get_tp_fp_fn_segment_wise(y_true: np.array,y_pred: np.array):
    

    y_true_anomaly_events = get_events(y_true)
    pred_anomaly_events = get_events(y_pred)
    y_true_normal_events = get_events(y_true,False)
    pred_normal_events = get_events(y_pred,False)

    TP = 0
    FN = 0
    FP = 0
    #TP
    i = 0
    for e_p in pred_anomaly_events:
        
        c, d = e_p
        while i<len(y_true_anomaly_events):
            e_g = y_true_anomaly_events[i]
            a, b = e_g
            if a>d:
              break

            if b<c:
                i+=1
                continue

            else:
              if max(a, c) <= min(b, d):
                  TP+=1
            

            i+=1

    #FN
    FN = len(y_true_anomaly_events) - TP
    #FP
    i = 0
    for e_p in y_true_normal_events:
        
        c, d = e_p
        while i<len(pred_anomaly_events):
            e_g = pred_anomaly_events[i]
            a, b = e_g
            if a>d:
                break
            if b<c:
                i+=1
                continue
            if calculate_intersection(e_g, e_p) is not None:
                FP+=1
            i+=1
    return TP, FP, FN

def is_full_series(length: int, anomalies: np.array):
    # [1 0 1 1 0]
    return len(anomalies.shape) == 1 and len(anomalies) == length

def is_pointwise(length: int, anomalies: np.array):
    # [0 2 3]
    return len(anomalies.shape) == 1 and len(anomalies) < length

def is_segmentwise(length: int, anomalies: np.array):
    # [[0 0] [2 3]]
    return len(anomalies.shape) == 2


def transform_to_full_series(length: int, anomalies: np.array):
    if is_full_series(length, anomalies):
        return anomalies
    elif is_pointwise(anomalies):
        return pointwise_to_full_series(anomalies, length)
    elif is_segmentwise(length, anomalies):
        return segmentwise_to_full_series(anomalies, length)
    else:
        raise ValueError(f"Illegal shape of anomalies:\n{anomalies}")

def counting_method(y_true: np.array, y_pred: np.array, k: int):
    em,da,ma,fa = 0,0,0,0
    for gt,pa in zip(y_true,y_pred):
        if gt==1 and pa==1:
            em+=1
        elif gt==0 and pa==1:
            fa+=1
        else:
            pass
    b = DelayThresholdedPointAdjust(len(y_true),y_true,y_pred,k=k)
    da = b.tp-em
    ma = b.fn

    return em,da,ma,fa


#Range Based utils

def cardinality(n_intersections,mode):
    if mode == 'one':
        return 1
    elif mode == 'reciprocal':
        if n_intersections==0:
            return 1
        else:
            return float(1/n_intersections)
    else:
        raise Exception("Error, wrong cardinality mode.")
    
    
def size(anomaly_range, overlap_set, position, bias):
    if overlap_set == None:
        return 0
    
    my_value = 0
    max_value = 0
    anomaly_length = anomaly_range[1] - anomaly_range[0] + 1
    for i in range(1,anomaly_length+1):
        bias_value = position(i, anomaly_length,bias)
        max_value += bias_value
        if  anomaly_range[0]+i-1 >= overlap_set[0] and anomaly_range[0]+i-1 <= overlap_set[1]:
            my_value += bias_value
    return my_value / max_value

def position(i, anomaly_length,bias):
    if bias == 'flat':
        return 1
    elif bias == 'front-end':
        return anomaly_length - i + 1
    elif bias == 'back-end':
        return i
    elif bias == 'middle':
        if i <= anomaly_length / 2:
            return i
        else:
            return anomaly_length - i + 1
    else:
        raise Exception("Error, wrong bias value.")


import unittest
from tsadmetrics import *
import numpy as np
import random


class TestPrecisionAtK(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
 
        self.y_true1 =  np.array([0,0,1,1])


        self.y_pred1 = np.array([0.2, 0.9, 0.3, 0.8])

        self.y_pred2 = np.array([1, 2, 3, 4])

        self.y_pred3 = np.array([3, 4, 1, 2])

        self.y_true2 =  np.array([1,1,1,0])

        self.y_pred4 = np.array([3, 4, 1, 2])


    

    def test_precision_at_k_score(self):
        """
        Prueba para la función precision_at_k_score.
        """
        score = round(precision_at_k(self.y_true1, self.y_pred1),2)
        expected_score = 0.5
        self.assertAlmostEqual(score, expected_score, places=4)

        score = round(precision_at_k(self.y_true1, self.y_pred2),2)
        expected_score = 1
        self.assertAlmostEqual(score, expected_score, places=4)

        score = round(precision_at_k(self.y_true1, self.y_pred3),2)
        expected_score = 0
        self.assertAlmostEqual(score, expected_score, places=4)

        score = round(precision_at_k(self.y_true2, self.y_pred4),2)
        expected_score = round(2/3,2)
        self.assertAlmostEqual(score, expected_score, places=4)
        
    def test_precision_at_k_score_consistency(self):
        try:
            for _ in range(100):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.random( size=(100,))

                score = precision_at_k(y_true, y_pred)
        except Exception as e:
            self.fail(f"precision_at_k_score raised an exception {e}")


class TestAUCROCPW(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
 
        self.y_true1 =  np.array([0,0,1,1])


        self.y_pred1 = np.array([1, 3, 2, 4])

        self.y_pred2 = np.array([1, 2, 3, 4])

        self.y_pred3 = np.array([4, 4, 4, 4])
    

    def test_auc_roc_pw(self):
        """
        Prueba para la función auc_roc_pw.
        """
        score = round(auc_roc_pw(self.y_true1, self.y_pred1),2)
        expected_score = 0.75
        self.assertAlmostEqual(score, expected_score, places=4)

        score = round(auc_roc_pw(self.y_true1, self.y_pred2),2)
        expected_score = 1
        self.assertAlmostEqual(score, expected_score, places=4)

        score = round(auc_roc_pw(self.y_true1, self.y_pred3),2)
        expected_score = 0.5
        self.assertAlmostEqual(score, expected_score, places=4)

        
    def test_auc_roc_pw_consistency(self):
        try:
            for _ in range(100):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.random( size=(100,))

                score = auc_roc_pw(y_true, y_pred)
        except Exception as e:
            self.fail(f"auc_roc_pw raised an exception {e}")

class TestAUCPRPW(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
 
        self.y_true1 =  np.array([0,0,1,1])


        self.y_pred1 = np.array([1, 3, 2, 4])

        self.y_pred2 = np.array([1, 2, 3, 4])

        self.y_pred3 = np.array([4, 4, 4, 4])
    

    def test_auc_pr_pw(self):
        """
        Prueba para la función auc_pr_pw.
        """
        score = round(auc_pr_pw(self.y_true1, self.y_pred1),2)
        expected_score = 0.83
        self.assertAlmostEqual(score, expected_score, places=4)

        score = round(auc_pr_pw(self.y_true1, self.y_pred2),2)
        expected_score = 1
        self.assertAlmostEqual(score, expected_score, places=4)

        score = round(auc_pr_pw(self.y_true1, self.y_pred3),2)
        expected_score = 0.5
        self.assertAlmostEqual(score, expected_score, places=4)

        
    def test_auc_pr_consistency(self):
        try:
            for _ in range(100):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.random( size=(100,))

                score = auc_pr_pw(y_true, y_pred)
        except Exception as e:
            self.fail(f"auc_pr raised an exception {e}")


    

class TestAUCPRPA(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
 
        self.y_true1 =  np.array([0,0,1,1])


        self.y_pred1 = np.array([1, 3, 2, 4])

        self.y_pred2 = np.array([1, 2, 3, 4])

        self.y_pred3 = np.array([4, 4, 4, 4])

        self.y_true2 = [0,1,1,0,0,0,0,0,1,1,0,0,0,0,1,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,1,0,0,1,1,0
,1,1,1,0,0,1,0,0,1,0,1,1,0,0,1,0,0,0,0,1,0,0,0,0,1,0,0,1,0,1,1,1,1,1,0,1,1
,1,1,1,1,0,0,1,1,1,1,0,1,0,0,1,1,1,0,0,1,0,0,1,0,1,1]


        self.y_pred4 = [0.1280475, 0.12059283 ,0.29936968 ,0.85866402 ,0.74071874 ,0.22310849
 ,0.11281839 ,0.26133246 ,0.33696106 ,0.01442675 ,0.51962876 ,0.07828833
 ,0.45337844 ,0.09444483 ,0.91216588 ,0.18847595 ,0.26828481 ,0.65248919
 ,0.46291981 ,0.43730757 ,0.78087553 ,0.45031043 ,0.88661033 ,0.56209352
 ,0.45029423 ,0.17638205 ,0.9261279 ,0.58830652 ,0.01602648 ,0.73903379
 ,0.61831379 ,0.74779903 ,0.42682106 ,0.82583519 ,0.19709012 ,0.44925962
 ,0.62752415 ,0.52458327 ,0.46291768 ,0.33937527 ,0.34868777 ,0.12293847
 ,0.84477504 ,0.10225254 ,0.37048167 ,0.04476031 ,0.36680499 ,0.11346155
 ,0.10583112 ,0.09493136 ,0.54878736 ,0.68514489 ,0.5940307 ,0.14526962
 ,0.69385728 ,0.38888727 ,0.61495304 ,0.06795402 ,0.02894603 ,0.08293609
 ,0.22865685 ,0.63531487 ,0.97966126 ,0.31418622 ,0.8943095 ,0.22974177
 ,0.94402929 ,0.13140625 ,0.80539267 ,0.40160344 ,0.38151339 ,0.65011626
 ,0.71657942 ,0.93297398 ,0.32043329 ,0.54667941 ,0.90645979 ,0.98730183
 ,0.82351336 ,0.10404812 ,0.6962921 ,0.72890752 ,0.49700666 ,0.47461103
 ,0.59696079 ,0.85876179 ,0.247344 ,0.38187879 ,0.23906861 ,0.5266315
 ,0.08171512 ,0.27903375 ,0.61112439 ,0.20784267 ,0.90652453 ,0.87575255
 ,0.26972245 ,0.78780138 ,0.37649185 ,0.08467683]


    def test_auc_pr_pa(self):
        """
        Prueba para la función auc_pr_pa.
        """
        score = round(auc_pr_pa(self.y_true1, self.y_pred1),2)
        expected_score = 1.0
        self.assertAlmostEqual(score, expected_score, places=4)

        score = round(auc_pr_pa(self.y_true1, self.y_pred2),2)
        expected_score = 1.0
        self.assertAlmostEqual(score, expected_score, places=4)

        score = round(auc_pr_pa(self.y_true1, self.y_pred3),2)
        expected_score = 0.5
        self.assertAlmostEqual(score, expected_score, places=4)

        if len(self.y_true2) == len(self.y_pred4):
            score = round(auc_pr_pa(self.y_true2, self.y_pred4),2)
            expected_score = 0.78
            self.assertAlmostEqual(score, expected_score, places=4)

        
    def test_auc_pr_pa_consistency(self):
        y_true, y_pred = [],[]
        try:
            for _ in range(100):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.random( size=(100,))
                score = auc_pr_pa(y_true, y_pred)
        except Exception as e:
            self.fail(f"auc_roc_pr_pa raised an exception {e}")
            

class TestAUCPRSW(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
 
        self.y_true1 =  np.array([0,0,1,1])


        self.y_pred1 = np.array([1, 3, 2, 4])

        self.y_pred2 = np.array([1, 2, 3, 4])

        self.y_pred3 = np.array([4, 4, 4, 4])
    

    def test_auc_pr_sw(self):
        """
        Prueba para la función auc_pr_sw.
        """
        score = round(auc_pr_sw(self.y_true1, self.y_pred1),2)
        expected_score = 1.0
        self.assertAlmostEqual(score, expected_score, places=4)

        score = round(auc_pr_sw(self.y_true1, self.y_pred2),2)
        expected_score = 1
        self.assertAlmostEqual(score, expected_score, places=4)

        score = round(auc_pr_sw(self.y_true1, self.y_pred3),2)
        expected_score = 0.5
        self.assertAlmostEqual(score, expected_score, places=4)

        
    # def test_auc_pr_sw_consistency(self):
    #     try:
    #         for _ in range(100):
    #             y_true = np.random.choice([0, 1], size=(100,))
    #             y_pred = np.random.random( size=(100,))

    #             score = auc_pr_sw(y_true, y_pred)
    #     except Exception as e:
    #         self.fail(f"auc_pr_sw raised an exception {e}")


class TestVUSROC(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
 
        self.y_true1 =  np.array([1, 1, 0, 0, 0, 0, 0, 0, 0, 0])
        self.y_true2 =  np.array([0, 1, 0, 1, 0, 0, 0, 0, 0, 0])

        self.y_pred1 = np.array( [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])

        self.y_pred2 = np.array([8, 0, 9, 1, 7, 2, 3, 4, 5, 6])


    

    def test_vus_roc(self):
        """
        Prueba para la función vus_roc.
        """
        score = round(vus_roc(self.y_true1, self.y_pred1,window=4),2)
        self.assertTrue(score <= 0.1)

        score = round(vus_roc(self.y_true2, self.y_pred2,window=4),2)
        self.assertTrue(score > 0.4)

        score = vus_roc(self.y_true2, self.y_pred2,window=0)
        self.assertTrue(score < 0.4)

        
    def test_vus_roc_consistency(self):
        try:
            for _ in range(10):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.random( size=(100,))
                score = vus_roc(y_true, y_pred, window=4)
        except Exception as e:
            self.fail(f"auc_roc raised an exception {e}")


class TestVUSPR(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
 
        self.y_true1 =  np.array([1, 1, 0, 0, 0, 0, 0, 0, 0, 0])
        self.y_true2 =  np.array([0, 1, 0, 1, 0, 0, 0, 0, 0, 0])

        self.y_pred1 = np.array( [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])

        self.y_pred2 = np.array([8, 0, 9, 1, 7, 2, 3, 4, 5, 6])


    

    def test_vus_pr(self):
        """
        Prueba para la función vus_pr.
        """
        score = round(vus_pr(self.y_true1, self.y_pred1,window=4),2)
        self.assertTrue(score <= 0.2)

        score = round(vus_pr(self.y_true2, self.y_pred2,window=4),2)
        self.assertTrue(score > 0.5)

        score = vus_pr(self.y_true2, self.y_pred2,window=0)
        self.assertTrue(score < 0.5)

        
    def test_vus_pr_consistency(self):
        try:
            for _ in range(10):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.random( size=(100,))
                score = vus_pr(y_true, y_pred, window=4)
        except Exception as e:
            self.fail(f"auc_roc raised an exception {e}")


class TestNonBinaryPATE(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
        pass

    

        
    def test_pate_consistency(self):
        try:
            for _ in range(10):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.choice([0, 1], size=(100,))

                score = real_pate(y_true, y_pred, early=5, delay=5)
        except Exception as e:
            self.fail(f"real_pate raised an exception {e}")


import unittest
from tsadmetrics import *

from sklearn.metrics import recall_score, precision_score, fbeta_score
import numpy as np
import random

class TestPointWiseMetrics(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
        self.num_tests = 100  # Número de conjuntos de datos aleatorios a generar para las pruebas
        self.test_cases = []
        for _ in range(self.num_tests):
            y_true = np.random.choice([0, 1], size=(10000,))
            y_pred = np.random.choice([0, 1], size=(10000,))
            self.test_cases.append((y_true, y_pred))

    def test_point_wise_recall(self):
        """
        Prueba para la función point_wise_recall.
        """
        for y_true, y_pred in self.test_cases:
            with self.subTest(y_true=y_true, y_pred=y_pred):
                recall = point_wise_recall(y_true, y_pred)
                expected_recall = recall_score(y_true, y_pred)
                self.assertAlmostEqual(recall, expected_recall, places=4)

    def test_point_wise_precision(self):
        """
        Prueba para la función point_wise_precision.
        """
        for y_true, y_pred in self.test_cases:
            with self.subTest(y_true=y_true, y_pred=y_pred):
                precision = point_wise_precision(y_true, y_pred)
                expected_precision = precision_score(y_true, y_pred)
                self.assertAlmostEqual(precision, expected_precision, places=4)

    def test_point_wise_f_score(self):
        """
        Prueba para la función point_wise_f_score.
        """
        for y_true, y_pred in self.test_cases:
            with self.subTest(y_true=y_true, y_pred=y_pred):
                beta = random.randint(0,1000000)
                f_score = point_wise_f_score(y_true, y_pred, beta=1)
                expected_f_score = fbeta_score(y_true, y_pred, beta=1)
                self.assertAlmostEqual(f_score, expected_f_score, places=4)

class TestPointAdjustedMetrics(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
        self.y_true = np.array([0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0])
        self.y_pred = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0])

    def test_point_adjusted_recall(self):
        """
        Prueba para la función point_wise_recall.
        """
        recall = point_adjusted_recall(self.y_true, self.y_pred)
        expected_recall = 1
        self.assertAlmostEqual(recall, expected_recall, places=4)
    def test_point_adjusted_precision(self):
        """
        Prueba para la función point_adjusted_precision.
        """
        precision = round(point_adjusted_precision(self.y_true, self.y_pred),2)
        expected_precision = 0.87
        self.assertAlmostEqual(precision, expected_precision, places=4)

    def test_point_adjusted_f_score(self):
        """
        Prueba para la función point_adjusted_f_score.
        """
        f_score = round(point_adjusted_f_score(self.y_true, self.y_pred),2)
        expected_f_score = 0.93
        self.assertAlmostEqual(f_score, expected_f_score, places=4)
        
    def test_point_adjusted_consistency(self):
        try:
            y_true = np.random.choice([0, 1], size=(100,))
            y_pred = np.zeros(100)
            point_adjusted_f_score(y_true, y_pred)
            for _ in range(1000):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.choice([0, 1], size=(100,))
                f_score = point_adjusted_f_score(y_true, y_pred)
        except Exception as e:
            self.fail(f"point_adjusted_f_score raised an exception {e}")

class TestDelayThPointAdjustedMetrics(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
        self.y_true  = np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1])
        self.y_pred1 = np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])
        self.y_pred2 = np.array([0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0])

    

    def test_delay_th_point_adjusted_f_score(self):
        """
        Prueba para la función delay_th_point_adjusted_f_score.
        """
        f_score = round(delay_th_point_adjusted_f_score(self.y_true, self.y_pred1, 2),2)
        expected_f_score = 0.67
        self.assertAlmostEqual(f_score, expected_f_score, places=4)
        
        f_score = round(delay_th_point_adjusted_f_score(self.y_true, self.y_pred2, 2),2)
        expected_f_score = 1
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

    def test_delay_th_point_adjusted_consistency(self):
        try:
            y_true = np.random.choice([0, 1], size=(100,))
            y_pred = np.zeros(100)
            delay_th_point_adjusted_f_score(y_true, y_pred,7)
            for _ in range(1000):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.choice([0, 1], size=(100,))
                f_score = delay_th_point_adjusted_f_score(y_true, y_pred, 7)
        except Exception as e:
            self.fail(f"delay_th_point_adjusted_f_score raised an exception {e}")

class TestPointAdjustedMetricsAtK(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
        self.y_true  = np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1])
        self.y_pred1 = np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])
        self.y_pred2 = np.array([0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0])


    def test_point_adjusted_at_k_f_score(self):
        """
        Prueba para la función point_adjusted_at_k_f_score.
        """
        f_score = round(point_adjusted_at_k_f_score(self.y_true, self.y_pred1,0.2),2)
        expected_f_score = 0.67
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

        f_score = round(point_adjusted_at_k_f_score(self.y_true, self.y_pred2,0.2),2)
        expected_f_score = 0.22
        self.assertAlmostEqual(f_score, expected_f_score, places=4)
        
    def test_point_adjusted_at_k_consistency(self):
        try:
            y_true = np.random.choice([0, 1], size=(100,))
            y_pred = np.zeros(100)
            point_adjusted_at_k_f_score(y_true, y_pred,0.3)
            for _ in range(1000):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.choice([0, 1], size=(100,))
                f_score = point_adjusted_at_k_f_score(y_true, y_pred,0.3)
        except Exception as e:
            self.fail(f"point_adjusted_at_k_f_score raised an exception {e}")

class TestLatencySparsityAwareMetrics(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
        self.y_true  = np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1])
        self.y_pred1 = np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])
        self.y_pred2 = np.array([0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0])

    def test_latency_sparsity_aw_f_score(self):
        """
        Prueba para la función latency_sparsity_aw_f_score.
        """
        f_score = round(latency_sparsity_aw_f_score(self.y_true, self.y_pred1,2),2)
        expected_f_score = 0.71
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

        f_score = round(latency_sparsity_aw_f_score(self.y_true, self.y_pred2,2),2)
        expected_f_score = 1
        self.assertAlmostEqual(f_score, expected_f_score, places=4)
        
    def test_latency_sparsity_aw_consistency(self):
        try:
            y_true = np.random.choice([0, 1], size=(100,))
            y_pred = np.zeros(100)
            latency_sparsity_aw_f_score(y_true, y_pred,3)
            for _ in range(1000):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.choice([0, 1], size=(100,))
                f_score = latency_sparsity_aw_f_score(y_true, y_pred,3)
        except Exception as e:
            self.fail(f"latency_sparsity_aw_f_score raised an exception {e}")

class TestSegmentWiseMetrics(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
        self.y_true  = np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1])
        self.y_pred1 = np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])
        self.y_pred2 = np.array([0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0])


    def test_segment_wise_f_score(self):
        """
        Prueba para la función segment_wise_f_score.
        """
        f_score = round(segment_wise_f_score(self.y_true, self.y_pred1),2)
        expected_f_score = 0.67
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

        f_score = round(segment_wise_f_score(self.y_true, self.y_pred2),2)
        expected_f_score = 1
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

    
        
    def test_segment_wise_consistency(self):
        try:
            y_true = np.random.choice([0, 1], size=(100,))
            y_pred = np.zeros(100)
            segment_wise_f_score(y_true, y_pred,7)
            for _ in range(1000):
                y_true = np.random.choice([0, 1], size=(10,))
                y_pred = np.random.choice([0, 1], size=(10,))
                f_score = segment_wise_f_score(y_true, y_pred)
                
        except Exception as e:
            self.fail(f"segment_wise_f_score raised an exception {e}")

class TestCompositeMetrics(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
        self.y_true  = np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1])
        self.y_pred1 = np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])
        self.y_pred2 = np.array([0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0])




    def test_composite_f_score(self):
        """
        Prueba para la función composite_f_score.
        """
        f_score = round(composite_f_score(self.y_true, self.y_pred1),2)
        expected_f_score = 0.67
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

        f_score = round(composite_f_score(self.y_true, self.y_pred2),2)
        expected_f_score = 1
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

        
    def test_composite_consistency(self):
        try:
            y_true = np.random.choice([0, 1], size=(100,))
            y_pred = np.zeros(100)
            composite_f_score(y_true, y_pred,7)
            for _ in range(1000):
                y_true = np.random.choice([0, 1], size=(10,))
                y_pred = np.random.choice([0, 1], size=(10,))
                f_score = composite_f_score(y_true, y_pred)
                
        except Exception as e:
            self.fail(f"composite_f_score raised an exception {e}")

class TestTimeTolerantMetrics(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
        self.y_true =   np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1])
        self.y_pred1 =  np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])
        self.y_pred2 =  np.array([0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0])

    def test_time_tolerant_recall(self):
        """
        Prueba para la función time_tolerant_recall.
        """
        recall = round(time_tolerant_recall(self.y_true, self.y_pred1,2),2)
        expected_recall = 0.5
        self.assertAlmostEqual(recall, expected_recall, places=4)

        recall = round(time_tolerant_recall(self.y_true, self.y_pred2,2),3)
        expected_recall = 0.375
        self.assertAlmostEqual(recall, expected_recall, places=4)

    def test_time_tolerant_precision(self):
        """
        Prueba para la función time_tolerant_precision.
        """
        precision = round(time_tolerant_precision(self.y_true, self.y_pred1,2),2)
        expected_precision = 1
        self.assertAlmostEqual(precision, expected_precision, places=4)

        precision = round(time_tolerant_precision(self.y_true, self.y_pred2,2),2)
        expected_precision = 1
        self.assertAlmostEqual(precision, expected_precision, places=4)

    def test_time_tolerant_f_score(self):
        """
        Prueba para la función time_tolerant_f_score.
        """
        f_score = round(time_tolerant_f_score(self.y_true, self.y_pred1,2),2)
        expected_f_score = 0.67
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

        f_score = round(time_tolerant_f_score(self.y_true, self.y_pred2,2),2)
        expected_f_score = 0.55
        self.assertAlmostEqual(f_score, expected_f_score, places=4)
        
    def test_time_tolerant_consistency(self):
        try:
            y_true = np.random.choice([0, 1], size=(100,))
            y_pred = np.zeros(100)
            time_tolerant_f_score(y_true, y_pred,7)
            for _ in range(1000):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.choice([0, 1], size=(100,))
                t = random.randint(1,100)
                f_score = time_tolerant_f_score(y_true, y_pred,t)
        except Exception as e:
            self.fail(f"time_tolerant_f_score raised an exception {e}")


class TestRangeBasedMetrics(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
        self.y_true1 = np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1])
        self.y_pred1 = np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])
        self.y_pred2 = np.array([0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0])

        self.y_true2  = np.array([0,0,1,0,1,0,1,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0])
        self.y_pred21 = np.array([0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])
        self.y_pred22 = np.array([0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0])
    

    def test_range_based_f_score(self):
        """
        Prueba para la función range_based_f_score.
        """
        f_score = round(range_based_f_score(self.y_true1, self.y_pred1, beta=1,p_alpha=0.2,r_alpha=0.2,cardinality_mode='one',p_bias='flat',r_bias='flat'),2)
        expected_f_score = 0.67
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

        f_score = round(range_based_f_score(self.y_true1, self.y_pred2,beta=1,p_alpha=0.2,r_alpha=0.2,cardinality_mode='one',p_bias='flat',r_bias='flat'),2)
        expected_f_score = 0.46
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

        f_score = round(range_based_f_score(self.y_true2, self.y_pred21,beta=1,p_alpha=0.2,r_alpha=0.2,cardinality_mode='one',p_bias='flat',r_bias='flat'),2)
        expected_f_score = 0.71
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

        f_score = round(range_based_f_score(self.y_true2, self.y_pred22,beta=1,p_alpha=0.2,r_alpha=0.2,cardinality_mode='one',p_bias='flat',r_bias='flat'),2)
        expected_f_score = 0.67
        self.assertAlmostEqual(f_score, expected_f_score, places=4)
        
    def test_range_based_consistency(self):
        try:
            modes = ['flat','front','back','middle']
            modes_c = ['one','reciprocal']

            y_true = np.random.choice([0, 1], size=(100,))
            y_pred = np.zeros(100)
            range_based_f_score(y_true, y_pred,beta=2,p_alpha=random.random(),r_alpha=random.random(),cardinality_mode=random.choice(modes_c),p_bias=random.choice(modes),r_bias=random.choice(modes))
            for _ in range(100):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.choice([0, 1], size=(100,))
                f_score = range_based_f_score(y_true, y_pred,beta=2,p_alpha=random.random(),r_alpha=random.random(),cardinality_mode=random.choice(modes_c),p_bias=random.choice(modes),r_bias=random.choice(modes))
        except Exception as e:
            self.fail(f"range_based_f_score raised an exception {e}")

class TestTSAwareMetrics(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
        self.y_true1 = np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1])
        self.y_pred1 = np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])
        self.y_pred2 = np.array([0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0])

        self.y_true2 = np.array([0,0,1,0,1,0,1,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0])
        self.y_pred21 = np.array([0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])
        self.y_pred22 = np.array([0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0])

    

    def test_ts_aware_f_score(self):
        """
        Prueba para la función ts_aware_f_score.
        """
        f_score = round(ts_aware_f_score(self.y_true1, self.y_pred1,1, 0.5, 0, 0.5),2)
        expected_f_score = 0.67
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

        f_score = round(ts_aware_f_score(self.y_true1, self.y_pred2,1, 0.5, 0, 0.5),2)
        expected_f_score = 0.12
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

        f_score = round(ts_aware_f_score(self.y_true2, self.y_pred21,1, 0.5, 0, 0.5),2)
        expected_f_score = 0.77
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

        f_score = round(ts_aware_f_score(self.y_true2, self.y_pred22,1, 0.5, 0, 0.5),2)
        expected_f_score = 0.67
        self.assertAlmostEqual(f_score, expected_f_score, places=4)
        
    def test_ts_aware_consistency(self):
        try:
            y_true = np.random.choice([0, 1], size=(100,))
            y_pred = np.zeros(100)
            ts_aware_f_score(y_true, y_pred, 1, random.random(), 0, random.random())
            for _ in range(100):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.choice([0, 1], size=(100,))

                f_score = ts_aware_f_score(y_true, y_pred, 1, random.random(), 0, random.random())
        except Exception as e:
            self.fail(f"ts_aware_f_score raised an exception {e}")

class TestEnhancedTSAwareMetrics(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
        self.y_true1 = np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1])
        self.y_pred1 = np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])
        self.y_pred2 = np.array([0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0])

        self.y_true2 = np.array([0,0,1,0,1,0,1,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0])
        self.y_pred21 = np.array([0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])
        self.y_pred22 = np.array([0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0])

    

    def test_enhanced_ts_aware_f_score(self):
        """
        Prueba para la función ts_aware_f_score.
        """
        f_score = round(enhanced_ts_aware_f_score(self.y_true1, self.y_pred1,1, 0.5, 0.1),2)
        expected_f_score = 0.67
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

        f_score = round(enhanced_ts_aware_f_score(self.y_true1, self.y_pred2,1, 0.5, 0.1),2)
        expected_f_score = 0.72
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

        f_score = round(enhanced_ts_aware_f_score(self.y_true2, self.y_pred21,1, 0.5, 0.1),2)
        expected_f_score = 0.77
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

        f_score = round(enhanced_ts_aware_f_score(self.y_true2, self.y_pred22,1, 0.5, 0.1),2)
        expected_f_score = 0.67
        self.assertAlmostEqual(f_score, expected_f_score, places=4)
        
    def test_enhanced_ts_aware_consistency(self):
        try:
            y_true = np.random.choice([0, 1], size=(100,))
            y_pred = np.zeros(100)
            enhanced_ts_aware_f_score(y_true, y_pred, 1, random.random(), random.random())
            for _ in range(100):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.choice([0, 1], size=(100,))

                f_score = enhanced_ts_aware_f_score(y_true, y_pred, 1, random.random(), random.random())
        except Exception as e:
            self.fail(f"enhanced_ts_aware_f_score raised an exception {e}")


class TestAffiliationBasedMetrics(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
        self.y_true1 = np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1])
        self.y_pred1 = np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])
        self.y_pred2 = np.array([0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0])

        self.y_true2 = np.array([0,0,1,0,1,0,1,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0])
        self.y_pred21 = np.array([0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])
        self.y_pred22 = np.array([0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0])

    

    def test_affiliation_based_f_score(self):
        """
        Prueba para la función ts_aware_f_score.
        """
        f_score = round(affiliation_based_f_score(self.y_true1, self.y_pred1,1),2)
        expected_f_score = 0.67
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

        f_score = round(affiliation_based_f_score(self.y_true1, self.y_pred2,1),2)
        expected_f_score = 0.77
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

        f_score = round(affiliation_based_f_score(self.y_true2, self.y_pred21,1),2)
        expected_f_score = 0.77
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

        f_score = round(affiliation_based_f_score(self.y_true2, self.y_pred22,1),2)
        expected_f_score = 0.67
        self.assertAlmostEqual(f_score, expected_f_score, places=4)
        
    def test_affiliation_based_consistency(self):
        try:
            y_true = np.random.choice([0, 1], size=(100,))
            y_pred = np.zeros(100)
            affiliation_based_f_score(y_true, y_pred, 1)
            for _ in range(100):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.choice([0, 1], size=(100,))

                f_score = affiliation_based_f_score(y_true, y_pred, 1)
        except Exception as e:
            self.fail(f"affiliation_based_f_score raised an exception {e}")



class TestNABScore(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
        self.y_true1 = np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1])
        self.y_pred1 = np.array([0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])
        self.y_pred2 = np.array([0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0])

        self.y_true2  = np.array([0,0,0,1,1,0,0,1,1,0,0,0,0,0,0,0,0,1,1,1,1,1])
        self.y_pred21 = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1])
        self.y_pred22 = np.array([0,0,0,1,1,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0])

    

    def test_nab_score(self):
        """
        Prueba para la función ts_aware_f_score.
        """
        f_score = round(nab_score(self.y_true1, self.y_pred1),2)
        expected_f_score = 50
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

        f_score = round(nab_score(self.y_true1, self.y_pred2),2)
        expected_f_score = 100
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

        f_score = round(nab_score(self.y_true2, self.y_pred21),2)
        expected_f_score = 33.33
        self.assertAlmostEqual(f_score, expected_f_score, places=4)

        f_score = round(nab_score(self.y_true2, self.y_pred22),2)
        expected_f_score = 66.67
        self.assertAlmostEqual(f_score, expected_f_score, places=4)
        
    def test_nab_score_consistency(self):
        try:
            y_true = np.random.choice([0, 1], size=(100,))
            y_pred = np.zeros(100)
            nab_score(y_true, y_pred)
            for _ in range(100):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.choice([0, 1], size=(100,))

                score = nab_score(y_true, y_pred)
        except Exception as e:
            self.fail(f"nab_score raised an exception {e}")


class TestAverageDetectionCount(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
        pass

    

        
    def test_average_detection_count_consistency(self):
        try:
            y_true = np.random.choice([0, 1], size=(100,))
            y_pred = np.zeros(100)
            average_detection_count(y_true, y_pred)
            for _ in range(100):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.choice([0, 1], size=(100,))

                score = average_detection_count(y_true, y_pred)
        except Exception as e:
            self.fail(f"average_detection_count raised an exception {e}")

class TestAbsoluteDetectionDistance(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
        pass

    

        
    def test_absolute_detection_distance_consistency(self):
        try:
            y_true = np.random.choice([0, 1], size=(100,))
            y_pred = np.zeros(100)
            absolute_detection_distance(y_true, y_pred)
            for _ in range(100):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.choice([0, 1], size=(100,))

                score = absolute_detection_distance(y_true, y_pred)
        except Exception as e:
            self.fail(f"absolute_detection_distance raised an exception {e}")

class TestTotalDetectedInRange(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
        pass

    

        
    def test_total_detected_in_range_consistency(self):
        try:
            y_true = np.random.choice([0, 1], size=(100,))
            y_pred = np.zeros(100)
            total_detected_in_range(y_true, y_pred,k=4)
            for _ in range(100):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.choice([0, 1], size=(100,))

                score = total_detected_in_range(y_true, y_pred,k=4)
        except Exception as e:
            self.fail(f"total_detected_in_range raised an exception {e}")

class TestDetectionAccuracyInRange(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
        pass

    

        
    def test_detection_accuracy_in_range_consistency(self):
        try:
            y_true = np.random.choice([0, 1], size=(100,))
            y_pred = np.zeros(100)
            detection_accuracy_in_range(y_true, y_pred,k=4)
            for _ in range(100):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.choice([0, 1], size=(100,))

                score = detection_accuracy_in_range(y_true, y_pred,k=4)
        except Exception as e:
            self.fail(f"detection_accuracy_in_range raised an exception {e}")


class TestWeightedDetectionDifference(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
        pass

    

        
    def test_weighted_detection_difference_consistency(self):
        try:
            y_true = np.random.choice([0, 1], size=(100,))
            y_pred = np.zeros(100)
            weighted_detection_difference(y_true, y_pred,k=4)
            for _ in range(100):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.choice([0, 1], size=(100,))

                score = weighted_detection_difference(y_true, y_pred,k=4)
        except Exception as e:
            self.fail(f"weighted_detection_difference raised an exception {e}")

class TestPATE(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
        pass

    

        
    def test_pate_consistency(self):
        try:
            y_true = np.random.choice([0, 1], size=(100,))
            y_pred = np.zeros(100)
            binary_pate(y_true, y_pred, early=5, delay=5)
            for _ in range(10):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.choice([0, 1], size=(100,))

                score = binary_pate(y_true, y_pred, early=5, delay=5)
        except Exception as e:
            self.fail(f"binary_pate raised an exception {e}")


class TestMeanTimeToDetect(unittest.TestCase):

    def setUp(self):
        """
        Configuración inicial para las pruebas.
        """
        pass

    

        
    def test_mean_time_to_detect_consistency(self):
        try:
            y_true = np.random.choice([0, 1], size=(100,))
            y_pred = np.zeros(100)
            mean_time_to_detect(y_true, y_pred)
            for _ in range(100):
                y_true = np.random.choice([0, 1], size=(100,))
                y_pred = np.random.choice([0, 1], size=(100,))

                score = mean_time_to_detect(y_true, y_pred)
        except Exception as e:
            self.fail(f"mean_time_to_detect raised an exception {e}")

if __name__ == '__main__':
    unittest.main()



from .binary_metrics import *
from .non_binary_metrics import *
from .utils import *
__author__ = 'Pedro Rafael Velasco Priego i12veprp@uco.es'
__version__ = "0.0.1"
__all__ = ['point_wise_recall', 'point_wise_precision', 'point_wise_f_score','point_adjusted_recall',
            'point_adjusted_precision', 'point_adjusted_f_score', 'segment_wise_recall', 'segment_wise_precision',
            'segment_wise_f_score','delay_th_point_adjusted_recall', 'delay_th_point_adjusted_precision',
              'delay_th_point_adjusted_f_score','point_adjusted_at_k_recall','point_adjusted_at_k_precision',
              'point_adjusted_at_k_f_score','latency_sparsity_aw_recall', 'latency_sparsity_aw_precision',
                'latency_sparsity_aw_f_score','composite_f_score','time_tolerant_recall','time_tolerant_precision',
                'time_tolerant_f_score','range_based_recall','range_based_precision','range_based_f_score','ts_aware_recall',
                'ts_aware_precision','ts_aware_f_score','enhanced_ts_aware_recall','enhanced_ts_aware_precision','enhanced_ts_aware_f_score',
                'affiliation_based_recall','affiliation_based_precision','affiliation_based_f_score','nab_score','temporal_distance',
                'average_detection_count','absolute_detection_distance','total_detected_in_range','detection_accuracy_in_range','weighted_detection_difference',
                'binary_pate','real_pate','mean_time_to_detect',
                'precision_at_k','auc_roc_pw','auc_pr_pw','auc_pr_pa','auc_pr_sw','vus_roc','vus_pr', 'compute_metrics',]


import numpy as np
import pandas as pd

def compute_metrics(y_true: np.array,y_pred: np.array,metrics: list, metrics_params: dict, is_anomaly_score = False):
    """
    Computes the specified metrics for the given true and predicted values.

    Parameters:
    - y_true (np.array): True labels.
    - y_pred (np.array): Predicted labels or scores.
    - metrics (list): List of metric names to compute.
    - metrics_params (dict): Dictionary of parameters for each metric.
    - is_anomaly_score (bool): Flag indicating if y_true and y_pred are anomaly scores. Otherwise, they are treated as binary labels.

    Returns:
    - metrics_df (DataFrame): DataFrame containing the computed metrics and their values.
    """

    if not is_anomaly_score:
        #Chech if y_true and y_pred are binary labels
        if not (np.array_equal(np.unique(y_true), [0, 1]) and np.array_equal(np.unique(y_pred), [0, 1])):
            raise ValueError("y_true and y_pred must be binary labels (0 or 1) when is_anomaly_score is False. Which is the default.")
    else:
        # Check if y_true and y_pred are anomaly scores
        if not (np.array_equal(np.unique(y_true), [0, 1]) and np.array_equal(np.unique(y_pred), [0, 1])):
            raise ValueError("y_true and y_pred must be anomaly scores in range [0,1] when is_anomaly_score is True.")
    results = {}
    
    for metric in metrics:
        metric_name = metric[0]
        metric_func = metric[1]

        metric_value = metric_func(y_true, y_pred, **metrics_params.get(metric_name, {}))
        
        # Store the result in the DataFrame
        results[metric_name] = metric_value

    metrics_df = pd.DataFrame(columns=['metric_name', 'metric_value'])
    metrics_df['metric_name'] = results.keys()
    metrics_df['metric_value'] = results.values()
    
    return metrics_df


import numpy as np
from _tsadeval.metrics import *
from .metric_utils import transform_to_full_series
from sklearn.metrics import auc
from .binary_metrics import point_adjusted_precision, point_adjusted_recall, segment_wise_precision, segment_wise_recall
from pate.PATE_metric import PATE
def precision_at_k(y_true : np.array ,y_anomaly_scores:  np.array):
    
    m = PatK_pw(y_true,y_anomaly_scores)

    return m.get_score()

def auc_roc_pw(y_true : np.array ,y_anomaly_scores:  np.array):
    
    m = AUC_ROC(y_true,y_anomaly_scores)

    return m.get_score()


def auc_pr_pw(y_true : np.array ,y_anomaly_scores:  np.array):
    
    m = AUC_PR_pw(y_true,y_anomaly_scores)

    return m.get_score()



def auc_pr_pa(y_true: np.array, y_anomaly_scores: np.array):
    thresholds = np.unique(y_anomaly_scores)[::-1]  # Descending order
    precisions = [1]
    recalls = [0]
    for t in thresholds[:-1]:

        y_pred = (y_anomaly_scores >= t).astype(int)


        precisions.append(point_adjusted_precision(y_true, y_pred))
        recalls.append(point_adjusted_recall(y_true, y_pred))
    
    recalls.append(1)
    precisions.append(0)
    auc_value = auc(recalls, precisions)
    return auc_value




def auc_pr_sw(y_true: np.array, y_anomaly_scores: np.array):
    thresholds = np.unique(y_anomaly_scores)[::-1]  # Descending order
    precisions = [1]
    recalls = [0]
    
    for t in thresholds[:-1]:
        y_pred = (y_anomaly_scores >= t).astype(int)
        precisions.append(segment_wise_precision(y_true, y_pred))
        recalls.append(segment_wise_recall(y_true, y_pred))
    recalls.append(1)
    precisions.append(0)
    auc_value = auc(recalls, precisions)
    return auc_value


def vus_roc(y_true : np.array ,y_anomaly_scores:  np.array, window=4):
    
    m = VUS_ROC(y_true,y_anomaly_scores,max_window=window)

    return m.get_score()


def vus_pr(y_true : np.array ,y_anomaly_scores:  np.array,  window=4):
    
    m = VUS_PR(y_true,y_anomaly_scores,max_window=window)

    return m.get_score()


def real_pate(y_true: np.array, y_anomaly_scores: np.array, early: int, delay: int):
    """
    Calculate PATE score for anomaly detection in time series.
    The PATE score is the ratio of the number of true positives to the sum of true positives, false positives, and false negatives, within a given early and delay range.

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_anomaly_scores (np.array): The predicted binary labels for the time series data.
    early (int): The maximum number of time steps before an anomaly must be predicted to be considered early.
    delay (int): The maximum number of time steps after an anomaly must be predicted to be considered delayed.

    Returns:
    float: The PATE score.
    """
    
    return PATE(y_true, y_anomaly_scores, early, delay, binary_scores=False)


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="tsadmetrics",
    version="0.1.0",
    author="Pedro Rafael Velasco Priego",
    author_email="i12veprp@uco.es",
    description="A library for time series anomaly detection metrics and evaluation.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pathsko/TSADmetrics",
    packages=find_packages(where="src"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    python_requires=">=3.6",
    install_requires=[
        "joblib==1.4.2",
        "numpy==1.24.4",
        "pandas==2.0.3",
        "PATE==0.1.1",
        "patsy==0.5.6",
        "python-dateutil==2.9.0.post0",
        "pytz==2024.1",
        "scikit-learn==1.3.2",
        "scipy==1.10.1",
        "six==1.16.0",
        "statsmodels==0.14.1",
        "threadpoolctl==3.5.0",
        "tzdata==2024.1",
    ],
    extras_require={
        "dev": [
            "pytest>=6.0",
            "ipython>=7.0",
            "sphinx",
            "sphinx-rtd-theme",
            "numpydoc",
            "myst-parser",
        ],
    },
    package_dir={"": "src"},
)



import numpy as np
from .metric_utils import *
from .metric_utils import get_events, calculate_intersection


from _tsadeval.metrics import *
from _tsadeval.prts.basic_metrics_ts import ts_fscore
from pate.PATE_metric import PATE
def point_wise_recall(y_true: np.array, y_pred: np.array):
    """
    Calculate point-wise recall for anomaly detection in time series.
    Implementation of https://link.springer.com/article/10.1007/s10618-023-00988-8

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.

    Returns:
    float: The point-wise recall score, which is the ratio of true positives to the sum of true positives and false negatives.
    """
    m = Pointwise_metrics(len(y_true),y_true,y_pred)
    m.set_confusion()
    TP,FN = m.tp,m.fn
    #TP, _, FP, FN = get_tp_tn_fp_fn_point_wise(y_true, y_pred)
    if TP == 0:
        return 0
    return TP / (TP + FN)

def point_wise_precision(y_true: np.array, y_pred: np.array):
    """
    Calculate point-wise precision for anomaly detection in time series.
    Implementation of https://link.springer.com/article/10.1007/s10618-023-00988-8

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.

    Returns:
    float: The point-wise precision score, which is the ratio of true positives to the sum of true positives and false positives.
    """
    #TP, _, FP, FN = get_tp_tn_fp_fn_point_wise(y_true, y_pred)
    m = Pointwise_metrics(len(y_true),y_true,y_pred)
    m.set_confusion()
    TP,FP = m.tp,m.fp
    if TP == 0:
        return 0
    return TP / (TP + FP)

def point_wise_f_score(y_true: np.array, y_pred: np.array, beta=1):
    """
    Calculate point-wise F-score for anomaly detection in time series.
    Implementation of https://link.springer.com/article/10.1007/s10618-023-00988-8

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    betha (float): The beta value, which determines the weight of precision in the combined score.
                  Default is 1, which gives equal weight to precision and recall.

    Returns:
    float: The point-wise F-score, which is the harmonic mean of precision and recall, adjusted by the beta value.
    """
    precision = point_wise_precision(y_true, y_pred)
    recall = point_wise_recall(y_true, y_pred)

    if precision == 0 or recall == 0:
        return 0
    
    return ((1 + beta**2) * precision * recall) / (beta**2 * precision + recall)


def point_adjusted_recall(y_true: np.array, y_pred: np.array):
    """
    Calculate point-adjusted recall for anomaly detection in time series.
    Implementation of https://link.springer.com/article/10.1007/s10618-023-00988-8

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.

    Returns:
    float: The point-adjusted recall score, which is the ratio of true positives to the sum of true positives and false negatives.
    """
    if np.sum(y_pred) == 0:
        return 0
    m = PointAdjust(len(y_true),y_true,y_pred)
    #m.adjust()
    TP,FN = m.tp,m.fn
    if TP == 0:
        return 0
    return TP / (TP + FN)

def point_adjusted_precision(y_true: np.array, y_pred: np.array):
    """
    Calculate point-adjusted precision for anomaly detection in time series.
    Implementation of https://link.springer.com/article/10.1007/s10618-023-00988-8

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.

    Returns:
    float: The point-adjusted precision score, which is the ratio of true positives to the sum of true positives and false positives.
    """
    if np.sum(y_pred) == 0:
        return 0
    m = PointAdjust(len(y_true),y_true,y_pred)
    #m.adjust()
    TP,FP = m.tp,m.fp
    if TP == 0:
        return 0
    return TP / (TP + FP)

def point_adjusted_f_score(y_true: np.array, y_pred: np.array, beta=1):
    """
    Calculate point-adjusted F-score for anomaly detection in time series.
    Implementation of https://link.springer.com/article/10.1007/s10618-023-00988-8

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    beta (float): The beta value, which determines the weight of precision in the combined score.
                  Default is 1, which gives equal weight to precision and recall.

    Returns:
    float: The point-adjusted F-score, which is the harmonic mean of precision and recall, adjusted by the beta value.
    """
    precision = point_adjusted_precision(y_true, y_pred)
    recall = point_adjusted_recall(y_true, y_pred)

    if precision == 0 or recall == 0:
        return 0
    
    return ((1 + beta**2) * precision * recall) / (beta**2 * precision + recall)



def delay_th_point_adjusted_recall(y_true: np.array, y_pred: np.array, k: int):
    """
    Calculate delay thresholded point-adjusted recall for anomaly detection in time series.
    Implementation of https://link.springer.com/article/10.1007/s10618-023-00988-8

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    k (int): The maximum number of time steps within which an anomaly must be predicted to be considered detected.

    Returns:
    float: The delay thresholded point-adjusted recall score, which is the ratio of true positives to the sum of true positives and false negatives.
    """
    if np.sum(y_pred) == 0:
        return 0
    m = DelayThresholdedPointAdjust(len(y_true),y_true,y_pred,k=k)
    TP,FN = m.tp,m.fn
    if TP == 0:
        return 0
    return TP / (TP + FN)

def delay_th_point_adjusted_precision(y_true: np.array, y_pred: np.array, k: int):
    """
    Calculate delay thresholded point-adjusted precision for anomaly detection in time series.
    Implementation of https://link.springer.com/article/10.1007/s10618-023-00988-8

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    k (int): The maximum number of time steps within which an anomaly must be predicted to be considered detected.

    Returns:
    float: The delay thresholded point-adjusted precision score, which is the ratio of true positives to the sum of true positives and false positives.
    """
    if np.sum(y_pred) == 0:
        return 0
    m = DelayThresholdedPointAdjust(len(y_true),y_true,y_pred,k=k)
    TP,FP = m.tp,m.fp
    if TP == 0:
        return 0
    return TP / (TP + FP)

def delay_th_point_adjusted_f_score(y_true: np.array, y_pred: np.array, k: int, beta=1):
    """
    Calculate delay thresholded point-adjusted F-score for anomaly detection in time series.
    Implementation of https://link.springer.com/article/10.1007/s10618-023-00988-8

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    k (int): The maximum number of time steps within which an anomaly must be predicted to be considered detected.
    beta (float): The beta value, which determines the weight of precision in the combined score.
                  Default is 1, which gives equal weight to precision and recall.

    Returns:
    float: The delay thresholded point-adjusted F-score, which is the harmonic mean of precision and recall, adjusted by the beta value.
    """
    precision = delay_th_point_adjusted_precision(y_true, y_pred, k)
    recall = delay_th_point_adjusted_recall(y_true, y_pred, k)

    if precision == 0 or recall == 0:
        return 0
    
    return ((1 + beta**2) * precision * recall) / (beta**2 * precision + recall)


def point_adjusted_at_k_recall(y_true: np.array, y_pred: np.array, k: float):
    """
    Calculate k percent point-adjusted at K% recall for anomaly detection in time series.
    Implementation of https://link.springer.com/article/10.1007/s10618-023-00988-8

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    k (float): The minimum percentage of the anomaly that must be detected to consider the anomaly as detected.

    Returns:
    float: The point-adjusted recall score, which is the ratio of true positives to the sum of true positives and false negatives.
    """
    #TP, _, FP, FN = get_tp_tn_fp_fn_point_adjusted_at_k(y_true, y_pred, k)
    m = PointAdjustKPercent(len(y_true),y_true,y_pred,k=k)
    TP,FN = m.tp,m.fn
    if TP == 0:
        return 0
    return TP / (TP + FN)

def point_adjusted_at_k_precision(y_true: np.array, y_pred: np.array, k: float):
    """
    Calculate point-adjusted at K% precision for anomaly detection in time series.
    Implementation of https://link.springer.com/article/10.1007/s10618-023-00988-8

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    k (float): The minimum percentage of the anomaly that must be detected to consider the anomaly as detected.

    Returns:
    float: The point-adjusted precision score, which is the ratio of true positives to the sum of true positives and false positives.
    """
    #TP, _, FP, _ = get_tp_tn_fp_fn_point_adjusted_at_k(y_true, y_pred, k)
    m = PointAdjustKPercent(len(y_true),y_true,y_pred,k=k)
    TP,FP = m.tp,m.fp
    if TP == 0:
        return 0
    return TP / (TP + FP)

def point_adjusted_at_k_f_score(y_true: np.array, y_pred: np.array, k: float, beta=1):
    """
    Calculate point-adjusted at K% F-score for anomaly detection in time series.
    Implementation of https://link.springer.com/article/10.1007/s10618-023-00988-8

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    k (float): The minimum percentage of the anomaly that must be detected to consider the anomaly as detected.
    beta (float): The beta value, which determines the weight of precision in the combined score.
                  Default is 1, which gives equal weight to precision and recall.

    Returns:
    float: The point-adjusted F-score, which is the harmonic mean of precision and recall, adjusted by the beta value.
    """
    precision = point_adjusted_at_k_precision(y_true, y_pred, k)
    recall = point_adjusted_at_k_recall(y_true, y_pred, k)

    if precision == 0 or recall == 0:
        return 0
    
    return ((1 + beta**2) * precision * recall) / (beta**2 * precision + recall)


def latency_sparsity_aw_recall(y_true: np.array, y_pred: np.array, ni: int):
    """
    Calculate latency and sparsity aware recall for anomaly detection in time series.
    Implementation of https://dl.acm.org/doi/10.1145/3447548.3467174

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    ni (int): The batch size used in the implementation to handle latency and sparsity.

    Returns:
    float: The latency and sparsity aware recall score, which is the ratio of true positives to the sum of true positives and false negatives.
    """
    if np.sum(y_pred) == 0:
        return 0
    m = LatencySparsityAware(len(y_true),y_true,y_pred,tw=ni)
    TP,FN = m.tp, m.fn
    if TP == 0:
        return 0
    return TP / (TP + FN)

def latency_sparsity_aw_precision(y_true: np.array, y_pred: np.array, ni: int):
    """
    Calculate latency and sparsity aware precision for anomaly detection in time series.
    Implementation of https://dl.acm.org/doi/10.1145/3447548.3467174

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    ni (int): The batch size used in the implementation to handle latency and sparsity.

    Returns:
    float: The latency and sparsity aware precision score, which is the ratio of true positives to the sum of true positives and false positives.
    """
    if np.sum(y_pred) == 0:
        return 0
    m = LatencySparsityAware(len(y_true),y_true,y_pred,tw=ni)
    TP,FP = m.tp, m.fp
    if TP == 0:
        return 0
    return TP / (TP + FP)

def latency_sparsity_aw_f_score(y_true: np.array, y_pred: np.array, ni: int, beta=1):
    """
    Calculate latency and sparsity aware F-score for anomaly detection in time series.
    Implementation of https://dl.acm.org/doi/10.1145/3447548.3467174

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    ni (int): The batch size used in the implementation to handle latency and sparsity.
    beta (float): The beta value, which determines the weight of precision in the combined score.
                  Default is 1, which gives equal weight to precision and recall.

    Returns:
    float: The latency and sparsity aware F-score, which is the harmonic mean of precision and recall, adjusted by the beta value.
    """
    if np.sum(y_pred) == 0:
        return 0

    recall = latency_sparsity_aw_recall(y_true,y_pred,ni)
    precision = latency_sparsity_aw_precision(y_true,y_pred,ni)
    if precision == 0 or recall == 0:
        return 0
    return ((1 + beta**2) * precision * recall) / (beta**2 * precision + recall)

    
def segment_wise_recall(y_true: np.array, y_pred: np.array):
    """
    Calculate segment-wise recall for anomaly detection in time series.
    Implementation of https://arxiv.org/pdf/1802.04431

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.

    Returns:
    float: The segment-wise recall score, which is the ratio of true positives to the sum of true positives and false negatives.
    """
    #TP, _, FN = get_tp_fp_fn_segment_wise(y_true, y_pred)
    m = Segmentwise_metrics(len(y_true),y_true,y_pred)
    TP,FN = m.tp,m.fn
    if TP == 0:
        return 0
    return TP / (TP + FN)

def segment_wise_precision(y_true: np.array, y_pred: np.array):
    """
    Calculate segment-wise precision for anomaly detection in time series.
    Implementation of https://arxiv.org/pdf/1802.04431

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.

    Returns:
    float: The segment-wise precision score, which is the ratio of true positives to the sum of true positives and false positives.
    """
    #TP, FP, _ = get_tp_fp_fn_segment_wise(y_true, y_pred)
    m = Segmentwise_metrics(len(y_true),y_true,y_pred)
    TP,FP = m.tp,m.fp
    if TP == 0:
        return 0
    return TP / (TP + FP)

def segment_wise_f_score(y_true: np.array, y_pred: np.array, beta=1):
    """
    Calculate segment-wise F-score for anomaly detection in time series.
    Implementation of https://arxiv.org/pdf/1802.04431

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    beta (float): The beta value, which determines the weight of precision in the combined score.
                  Default is 1, which gives equal weight to precision and recall.

    Returns:
    float: The segment-wise F-score, which is the harmonic mean of precision and recall, adjusted by the beta value.
  
    """
    m = Segmentwise_metrics(len(y_true),y_true,y_pred)
    TP,FN,FP = m.tp,m.fn,m.fp
    if TP==0:
        return 0
    
    precision = TP / (TP + FP)
    recall = TP / (TP + FN)
    
    if precision == 0 or recall == 0:
        return 0
    return ((1 + beta**2) * precision * recall) / (beta**2 * precision + recall)

def composite_f_score(y_true: np.array, y_pred: np.array, beta=1):
    """
    Calculate composite F-score for anomaly detection in time series.
    Implementation of https://ieeexplore.ieee.org/document/9525836

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    beta (float): The beta value, which determines the weight of precision in the combined score.
                  Default is 1, which gives equal weight to precision and recall.

    Returns:
    float: The composite F-score, which is the harmonic mean of precision and recall, adjusted by the beta value.
  
    """
    m = Composite_f(len(y_true),y_true,y_pred)
    #Point wise precision
    precision =  m.precision()#point_wise_precision(y_true,y_pred)

    #Segment wise recall
    recall = m.recall()#segment_wise_recall(y_true,y_pred)

    if precision==0 or recall==0:
        return 0

    return ((1 + beta**2) * precision * recall) / (beta**2 * precision + recall)

def time_tolerant_recall(y_true: np.array, y_pred: np.array, t: int) -> float:
    """
    Calculate time tolerant recall for anomaly detection in time series.
    Implementation of https://arxiv.org/pdf/1802.04431

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    t (int): The time tolerance parameter

    Returns:
    float: The time tolerant recall score, which is the ratio of true positives to the sum of true positives and false negatives.
    """
    if np.sum(y_pred) == 0:
        return 0
    
    m = Time_Tolerant(len(y_true),y_true,y_pred,d=t)
    return m.recall()

def time_tolerant_precision(y_true: np.array, y_pred: np.array, t: int) -> float:
    """
    Calculate time tolerant precision for anomaly detection in time series.
    Implementation of https://arxiv.org/pdf/1802.04431

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    t (int): The time tolerance parameter

    Returns:
    float: The time tolerant precision score, which is the ratio of true positives to the sum of true positives and false positives.
    """
    if np.sum(y_pred) == 0:
        return 0
    m = Time_Tolerant(len(y_true),y_true,y_pred, d=t)
    return m.precision()


def time_tolerant_f_score(y_true: np.array, y_pred: np.array,t: int, beta=1):
    """
    Calculate time tolerant F-score for anomaly detection in time series.
    Implementation of https://arxiv.org/pdf/1802.04431

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    t (int): The time tolerance parameter
    beta (float): The beta value, which determines the weight of precision in the combined score.
                  Default is 1, which gives equal weight to precision and recall.
    
    Returns:
    float: The time tolerant F-score, which is the harmonic mean of precision and recall, adjusted by the beta value.
  
    """
    precision = time_tolerant_precision(y_true,y_pred,t)
    recall = time_tolerant_recall(y_true,y_pred,t)
    if precision==0 or recall==0:
        return 0
    return ((1 + beta**2) * precision * recall) / (beta**2 * precision + recall)


def range_based_recall(y_true: np.array, y_pred: np.array, alpha: float, bias='flat', cardinality_mode='one'):
    """
    Calculate range-based recall for anomaly detection in time series.
    
    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    alpha (float): A parameter that controls the length of the range considered for true positives.
    bias (str): The type of bias to apply for weighting (flat, front-end, back-end, middle).
    cardinality (str, optional): ["one", "reciprocal", "udf_gamma"]. Defaults to "one".
    
    Returns:
    float: The range-based recall score.
    """
    if np.sum(y_pred) == 0:
        return 0
    m = Range_PR(len(y_true),y_true,y_pred,cardinality=cardinality_mode, alpha=alpha,bias=bias)
    return m.recall()



def range_based_precision(y_true: np.array, y_pred: np.array, alpha: float, bias='flat', cardinality_mode='one'):
    """
    Calculate range-based precision for anomaly detection in time series.
    
    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    alpha (float): A parameter that controls the length of the range considered for true positives.
    bias (str): The type of bias to apply for weighting (flat, front-end, back-end, middle).
    cardinality (str, optional): ["one", "reciprocal", "udf_gamma"]. Defaults to "one".
    
    Returns:
    float: The range-based precision score.
    """
    if np.sum(y_pred) == 0:
        return 0
    m = Range_PR(len(y_true),y_true,y_pred,cardinality=cardinality_mode, alpha=alpha,bias=bias)
    return m.precision()
    
    
    



def range_based_f_score(y_true: np.array, y_pred: np.array, p_alpha: float, r_alpha: float,  p_bias='flat', r_bias='flat', cardinality_mode='one', beta=1) -> float:
    """
    Calculate range-based F-score for anomaly detection in time series.

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    alpha (float): A parameter that controls the length of the range considered for true positives.
    p_bias: string, default="flat"
            Positional bias for precision. This should be "flat", "front", "middle", or "back"
    r_bias: string, default="flat"
        Positional bias for recall. This should be "flat", "front", "middle", or "back"
    cardinality_mode (str, optional): ["one", "reciprocal", "udf_gamma"]. Defaults to "one".
    beta (float): The beta value, which determines the weight of precision in the combined score.
                  Default is 1, which gives equal weight to precision and recall.

    Returns:
    float: The range-based F-score, which is the harmonic mean of precision and recall, adjusted by the beta value.
    """
    if np.sum(y_pred) == 0:
        return 0
    f = ts_fscore(y_true, y_pred, beta=beta, p_alpha=p_alpha, r_alpha=r_alpha, cardinality=cardinality_mode, p_bias=p_bias, r_bias=r_bias)
    return f




def ts_aware_recall(y_true: np.array, y_pred: np.array, alpha: float, delta: float, theta: float, past_range: bool = False):
    """
    Calculate time series aware recall for anomaly detection in time series.
    
    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    alpha (float): A parameter that controls the length of the range considered for true positives.
    past_range: Determines if the range is considered in the past or future as specified in https://www.mdpi.com/2079-9292/11/8/1213
    
    Returns:
    float: The time series aware recall score.
    """
    m = TaF(len(y_true),y_true,y_pred,alpha=alpha,theta=theta,delta=delta,past_range=past_range)
    return m.recall()




def ts_aware_precision(y_true: np.array, y_pred: np.array,alpha: float, delta: float, theta: float, past_range: bool = False):
    """
    Calculate time series aware precision for anomaly detection in time series.
    
    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    alpha (float): A parameter that controls the length of the range considered for true positives.
    past_range: Determines if the range is considered in the past or future as specified in https://www.mdpi.com/2079-9292/11/8/1213
    
    Returns:
    float: The time series aware precision score.
    """
    m = TaF(len(y_true),y_true,y_pred,alpha=alpha,theta=theta,delta=delta,past_range=past_range)
    return m.precision()

    



def ts_aware_f_score(y_true: np.array, y_pred: np.array, beta: float, alpha: float, delta: float, theta: float, past_range: bool = False):
    """
    Calculate time series aware F-score for anomaly detection in time series.

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    alpha (float): A parameter that controls the length of the range considered for true positives.
    beta (float): The beta value, which determines the weight of precision in the combined score.
                  Default is 1, which gives equal weight to precision and recall.
    past_range: Determines if the range is considered in the past or future as specified in https://www.mdpi.com/2079-9292/11/8/1213

    Returns:
    float: The time series aware F-score, which is the harmonic mean of precision and recall, adjusted by the beta value.
    """
    
    m = TaF(len(y_true),y_true,y_pred,alpha=alpha,theta=theta,delta=delta,past_range=past_range)
    precision = m.precision()
    recall = m.recall()
    if precision==0 or recall==0:
        return 0
    
    return ((1 + beta**2) * precision * recall) / (beta**2 * precision + recall)





def enhanced_ts_aware_recall(y_true: np.array, y_pred: np.array, theta: float):
    """
    Calculate enhanced time series aware recall for anomaly detection in time series.
    
    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    alpha (float): A parameter that controls the length of the range considered for true positives.
    
    Returns:
    float: The time series aware recall score.
    """
    if np.sum(y_pred) == 0:
        return 0
    m = eTaF(len(y_true),y_true,y_pred,theta_p=theta)
    return m.recall()




def enhanced_ts_aware_precision(y_true: np.array, y_pred: np.array, theta: float):
    """
    Calculate enhanced time series aware precision for anomaly detection in time series.
    
    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    alpha (float): A parameter that controls the length of the range considered for true positives.
    
    Returns:
    float: The time series aware precision score.
    """
    if np.sum(y_pred) == 0:
        return 0
    m = eTaF(len(y_true),y_true,y_pred,theta_p=theta)
    return m.precision()

    



def enhanced_ts_aware_f_score(y_true: np.array, y_pred: np.array, beta: float, theta_p: float, theta_r: float):
    """
    Calculate enhanced time series aware F-score for anomaly detection in time series.

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    beta (float): The beta value, which determines the weight of precision in the combined score.
                  Default is 1, which gives equal weight to precision and recall.

    Returns:
    float: The time series aware F-score, which is the harmonic mean of precision and recall, adjusted by the beta value.
    """
    if np.sum(y_pred) == 0:
        return 0
    m = eTaF(len(y_true),y_true,y_pred,theta_p=theta_p, theta_r=theta_r)
    return m.result['f1']



def affiliation_based_recall(y_true: np.array, y_pred: np.array):
    """
    Calculate affiliation based recall for anomaly detection in time series.

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    beta (float): The beta value, which determines the weight of precision in the combined score.
                  Default is 1, which gives equal weight to precision and recall.

    Returns:
    float: The time series aware F-score, which is the harmonic mean of precision and recall, adjusted by the beta value.
    """
    if np.sum(y_pred) == 0:
        return 0
    m = Affiliation(len(y_true),y_true,y_pred)
    s = m.get_score()
    return m.r


def affiliation_based_precision(y_true: np.array, y_pred: np.array):
    """
    Calculate affiliation based F-score for anomaly detection in time series.

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    beta (float): The beta value, which determines the weight of precision in the combined score.
                  Default is 1, which gives equal weight to precision and recall.

    Returns:
    float: The time series aware F-score, which is the harmonic mean of precision and recall, adjusted by the beta value.
    """
    if np.sum(y_pred) == 0:
        return 0
    m = Affiliation(len(y_true),y_true,y_pred)
    s = m.get_score()
    return m.p


def affiliation_based_f_score(y_true: np.array, y_pred: np.array, beta=1):
    """
    Calculate affiliation based F-score for anomaly detection in time series.

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    beta (float): The beta value, which determines the weight of precision in the combined score.
                  Default is 1, which gives equal weight to precision and recall.

    Returns:
    float: The time series aware F-score, which is the harmonic mean of precision and recall, adjusted by the beta value.
    """
    if np.sum(y_pred) == 0:
        return 0
    m = Affiliation(len(y_true),y_true,y_pred)
    return m.get_score(beta)


def nab_score(y_true: np.array, y_pred: np.array):
    """
    Calculate NAB score for anomaly detection in time series.

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.


    Returns:
    float: The nab score, which is the harmonic mean of precision and recall, adjusted by the beta value.
    """
    
    m = NAB_score(len(y_true),y_true,y_pred)
    return m.get_score()

def temporal_distance(y_true: np.array, y_pred: np.array, distance: int = 0):
    """
    Calculate temporal distane for anomaly detection in time series.

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    distance (int): The distance type parameter for the temporal distance calculation.
        0: Euclidean distance
        1: Squared Euclidean distance


    Returns:
    float: The temporal distance, which is the harmonic mean of precision and recall, adjusted by the beta value.
    """
    
    m = Temporal_Distance(len(y_true),y_true,y_pred,distance=distance)
    return m.get_score()

def average_detection_count(y_true: np.array, y_pred: np.array):
    """
    Calculate average detection count for anomaly detection in time series.

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.


    Returns:
    float: The average detection count.
    """
    
    b = Binary_detection(len(y_true),y_true,y_pred)
    azs = b.get_gt_anomalies_segmentwise()
    a_points = b.get_gt_anomalies_ptwise()

    counts = []
    for az in azs:
        count = 0
        for ap in a_points:
            if ap >= az[0] and ap <= az[1]:
                count+=1
        counts.append(count)
    
    return np.mean(counts)

def absolute_detection_distance(y_true: np.array, y_pred: np.array):
    """
    Calculate absolute detection distance for anomaly detection in time series.

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.


    Returns:
    float: The absolute detection distance.
    """
    
    b = Binary_detection(len(y_true),y_true,y_pred)
    azs = b.get_gt_anomalies_segmentwise()
    a_points = b.get_gt_anomalies_ptwise()

    distance = 0
    for az in azs:
        for ap in a_points:
            if ap >= az[0] and ap <= az[1]:
                center = int((az[0] + az[1]) / 2)
                distance+=abs(ap - center)
    
    return distance/len(a_points)


def total_detected_in_range(y_true: np.array, y_pred: np.array, k: int):
    """
    Calculate total detected in range for anomaly detection in time series.

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    k (int): The maximum number of time steps within which an anomaly must be predicted to be considered detected.

    Returns:
    float: The total detected in range.
    """
    if np.sum(y_pred) == 0:
        return 0
    em,da,ma,_ = counting_method(y_true, y_pred, k)

    
    return (em + da)/(em + da + ma)


def detection_accuracy_in_range(y_true: np.array, y_pred: np.array, k: int):
    """
    Calculate detection accuracy in range for anomaly detection in time series.

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    k (int): The maximum number of time steps within which an anomaly must be predicted to be considered detected.

    Returns:
    float: The total detected in range.
    """
    if np.sum(y_pred) == 0:
        return 0
    em,da,_,fa = counting_method(y_true, y_pred, k)

    
    return (em + da)/(em + da + fa)


def weighted_detection_difference(y_true: np.array, y_pred: np.array, k: int):
    """
    Calculate weighted detection difference for anomaly detection in time series.
    The weighted detection difference is the difference between the number of true positives and the number of false positives, weighted by the number of true positives.

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    k (int): The maximum number of time steps within which an anomaly must be predicted to be considered detected.

    Returns:
    float: The weighted detection difference.
    """
    if np.sum(y_pred) == 0:
        return 0
    
    def gaussian(dt,tmax):
        if dt < tmax:
            return 1- dt/tmax
        else:
            return -1

    tmax = len(y_true)
    
    ones_indices = np.where(y_true == 1)[0]
    
    y_modified = y_true.astype(float).copy()
    
    for i in range(len(y_true)):
        if y_true[i] == 0:
            dt = np.min(np.abs(ones_indices - i)) if len(ones_indices) > 0 else tmax
            y_modified[i] = gaussian(dt, tmax)
    
    ws = 0
    wf = 0
    for i in range(len(y_pred)):
        if y_pred[i] != 1:
            ws+=y_modified[i]
        else:
            wf+=y_modified[i]

    _,_,_,fa = counting_method(y_true, y_pred, k)

    
    return ws - wf*fa


def binary_pate(y_true: np.array, y_pred: np.array, early: int, delay: int):
    """
    Calculate PATE score for anomaly detection in time series.
    The PATE score is the ratio of the number of true positives to the sum of true positives, false positives, and false negatives, within a given early and delay range.

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.
    early (int): The maximum number of time steps before an anomaly must be predicted to be considered early.
    delay (int): The maximum number of time steps after an anomaly must be predicted to be considered delayed.

    Returns:
    float: The PATE score.
    """
    
    return PATE(y_true, y_pred, early, delay, binary_scores=True)

def mean_time_to_detect(y_true: np.array, y_pred: np.array):
    """
    Calculate mean time to detect for anomaly detection in time series.
    The mean time to detect is the average number of time steps between the ground truth anomaly and the predicted anomaly.

    Parameters:
    y_true (np.array): The ground truth binary labels for the time series data.
    y_pred (np.array): The predicted binary labels for the time series data.

    Returns:
    float: The mean time to detect.
    """
    
    b = Binary_detection(len(y_true),y_true,y_pred)
    a_events = b.get_gt_anomalies_segmentwise()
    t_sum = 0
    for _,b in a_events:
        for i in range(b,len(y_pred)):
            if y_pred[i] == 1:
                t_sum+=i-b
                break
    
    return t_sum/len(a_events)
